from dataclasses import dataclass, asdict, field, fields
from pathlib import Path

from fsspec import AbstractFileSystem

from typing import Sequence

from .api import (
    RsConnectApi,
    BaseEntity,
    User,
    Content,
    Bundle,
    RsConnectApiMissingContentError,
    RsConnectApiError,
    RsConnectApiResultError,
    RsConnectApiRequestError,
    RSC_CODE_OBJECT_DOES_NOT_EXIST,
)
from ..utils import isfilelike

# Misc ----


def _not_impl_args_kwargs(args, kwargs):
    return NotImplementedError(
        "Additional args and kwargs not supported." f"\nArgs: {args}\nKwargs: {kwargs}"
    )


# Bundles ----


@dataclass
class PinBundleManifestMetadata:
    appmode: str = "static"
    primary_rmd: "str|None" = None
    primary_html: str = "index.html"
    content_category: str = "pin"
    has_parameters: bool = False


@dataclass
class PinBundleManifest:
    version: int = 1
    local: str = "en_US"
    platform: str = "3.5.1"
    metadata: PinBundleManifestMetadata = field(
        default_factory=PinBundleManifestMetadata
    )
    packages: None = None
    files: list = field(default_factory=list)
    users: None = None

    @classmethod
    def from_directory(cls, dir_name, recursive: bool = True):
        root_dir = Path(dir_name)

        paths = root_dir.rglob("*") if recursive else root_dir.glob("*")
        flat_rel_files = [str(p.relative_to(root_dir)) for p in paths]

        return cls(files=flat_rel_files)

    @classmethod
    def add_manifest_to_directory(cls, dir_name: "str | Path", **kwargs) -> None:
        import json

        # TODO(question): R lib uses RSCONNECT_TAR env variable
        bundle = cls.from_directory(dir_name, **kwargs)
        with (Path(dir_name) / "manifest.json").open("w") as f_manifest:
            json.dump(bundle.to_dict(), f_manifest)

    def to_dict(self):
        return asdict(self)


# FSSPEC ----------------------------------------------------------------------


@dataclass
class EmptyPath:
    pass


@dataclass
class UserPath:
    username: str

    def path_to_field(self, field_name):
        all_fields = [field.name for field in fields(self)]
        keep_fields = all_fields[: all_fields.index(field_name) + 1]
        return "/".join(getattr(self, k) for k in keep_fields)


@dataclass
class ContentPath(UserPath):
    content: str


@dataclass
class BundlePath(ContentPath):
    bundle: str


@dataclass
class BundleFilePath(BundlePath):
    file_name: str


class RsConnectFs(AbstractFileSystem):
    protocol: str = "rsc"

    def __init__(self, server_url, **kwargs):
        if isinstance(server_url, RsConnectApi):
            self.api = server_url
        else:
            self.api = RsConnectApi(server_url, **kwargs)

        self._user_name_cache = {}
        self._content_name_cache = {}

    def ls(
        self, path, details=False, **kwargs
    ) -> "Sequence[BaseEntity] | Sequence[str]":
        """List contents of Rstudio Connect Server.

        Parameters
        ----------
        path: str
            "" -> users
            "<username>" -> user content
            "<username>/<content>" -> user content bundles
        """

        if isinstance(self.parse_path(path), EmptyPath):
            # root path specified, so list users
            all_results = self.api.get_users()

        else:
            entity = self.info(path)

            if isinstance(entity, User):
                all_results = self.api.get_content(entity["guid"])
            elif isinstance(entity, Content):
                all_results = self.api.get_content_bundles(entity["guid"])
            else:
                raise ValueError(
                    "path must have form {username} or {username}/{content_name}"
                )

        if not details:
            return [entry.get_name() for entry in all_results]

        return all_results

        # all_results = self.api.query(
        #    "applications/", filter="content_type:pin", count=1000
        # )

    def put(
        self,
        lpath,
        rpath,
        recursive=False,
        *args,
        access_type="acl",
        deploy=True,
        cls_manifest=PinBundleManifest,
        **kwargs,
    ) -> None:
        """Put a bundle onto Rstudio Connect.

        Parameters
        ----------
        lpath: str
            A path to the local bundle directory.
        rpath: str
            A path to the content where the bundle is being put.
        access_type:
            Who can use and view this content? Must be one of all, logged_in or acl.
        cls_manifest:
            If maniest does not exist, a class with an .add_manifest_to_directory()
            method.

        """

        parsed = self.parse_path(rpath)

        if len(args) or len(kwargs):
            raise _not_impl_args_kwargs(args, kwargs)

        if recursive is False:
            raise NotImplementedError(
                "Must set recursive to True in order to put any RSConnect content."
            )

        if not isinstance(parsed, ContentPath):
            # TODO: replace all these with a custom PathError
            raise ValueError("Path must point to content.")

        # Create content item if missing ----

        try:
            content = self.info(rpath)
        except RsConnectApiMissingContentError:
            # TODO: this could be seen as analogous to mkdir (which gets
            # called by pins anyway)
            content = self.api.post_content_item(parsed.content, access_type)

        # Create bundle (with manifest.json inserted if missing) ----

        if not (Path(lpath) / "manifest.json").exists():
            # TODO(question): does R pins copy content to tmp directory, or
            # insert mainfest.json into the source directory?
            cls_manifest.add_manifest_to_directory(lpath)

        bundle = self.api.post_content_bundle(content["guid"], lpath)

        # Deploy bundle ----

        if deploy:
            task = self.api.post_content_item_deploy(
                bundle["content_guid"], bundle["id"]
            )

            task = self.api.poll_tasks(task["task_id"])
            if task["code"] != 0 or not task["finished"]:
                raise RsConnectApiError(f"deployment failed for task: {task}")

        # TODO: should return bundle itself?
        return f"{rpath}/{bundle['id']}"

    def open(self, path: str, mode: str = "rb", *args, **kwargs):
        """Open a file inside an RStudio Connect bundle."""

        if mode != "rb":
            raise NotImplementedError()

        parsed = self.parse_path(path)

        if not isinstance(parsed, BundleFilePath):
            raise ValueError(
                "Path to a bundle file required. "
                "e.g. <user_name>/<content_name>/<bundle_id>/<file_name>"
            )

        bundle = self.info(
            parsed.path_to_field("bundle")
        )  # f"{parsed.username}/{parsed.content}/{parsed.bundle}")

        # TODO: do whatever other remote backends do
        from io import BytesIO

        f = BytesIO()

        self.api.misc_get_content_bundle_file(
            bundle["content_guid"], bundle["id"], parsed.file_name, f
        )
        f.seek(0)
        return f

    def get(self, rpath, lpath, recursive=False, *args, **kwargs) -> None:
        """Fetch a bundle or file from RStudio Connect."""
        parsed = self.parse_path(rpath)

        if recursive:
            if not isinstance(parsed, BundlePath):
                raise ValueError("Must receive path to bundle for recursive get.")

            bundle = self.info(rpath)
            self.api.get_content_bundle_archive(
                bundle["content_guid"], bundle["id"], lpath
            )

        elif isinstance(parsed, BundleFilePath):
            bundle = self.info(parsed.path_to_field("bundle"))
            self.api.misc_get_content_bundle_file(
                bundle["content_guid"], bundle["id"], parsed.file_name, lpath
            )

    def get_file(self, rpath, lpath, **kwargs):
        data = self.cat_file(rpath, **kwargs)
        if isfilelike(lpath):
            lpath.write(data)
        else:
            with open(lpath, "wb") as f:
                f.write(data)

    def exists(self, path: str, **kwargs) -> bool:
        try:
            self.info(path)
            return True
        except RsConnectApiMissingContentError:
            return False

    def mkdir(
        self, path, create_parents=True, *args, access_type="acl", **kwargs
    ) -> None:
        parsed = self.parse_path(path)

        if len(args) or len(kwargs):
            raise _not_impl_args_kwargs(args, kwargs)

        if not isinstance(parsed, ContentPath):
            raise ValueError(f"Requires path to content, but received: {path}")

        if self.exists(path):
            raise FileExistsError(path)

        # TODO: could implement and call makedirs, but seems overkill
        self.api.post_content_item(parsed.content, access_type, **kwargs)

    def info(self, path, **kwargs) -> "User | Content | Bundle":
        # TODO: source of fsspec info uses self._parent to check cache?
        # S3 encodes refresh (for local cache) and version_id arguments

        return self._get_entity_from_path(path)

    def rm(self, path, recursive=False, maxdepth=None) -> None:
        parsed = self.parse_path(path)

        # guards ----
        if maxdepth is not None:
            raise NotImplementedError("rm maxdepth argument not supported.")
        if isinstance(parsed, BundleFilePath):
            raise ValueError("Cannot rm a bundle file.")

        # time to delete things ----
        entity = self.info(path)

        if isinstance(entity, User):
            raise ValueError("Cannot rm a user.")
        if isinstance(entity, Content):
            if not recursive:
                raise ValueError("Must set recursive to true if deleting content.")

            self.api.delete_content_item(entity["guid"])

        elif isinstance(entity, Bundle):
            self.api.delete_content_bundle(entity["content_guid"], entity["id"])
        else:
            raise ValueError("Cannot entity: {type(entity)}")

    # Utils ----

    def parse_path(self, path):
        # root can be indicated by a slash
        if path.startswith("/"):
            path = path[1:]

        parts = path.split("/")
        if path.strip() == "":
            return EmptyPath()
        elif len(parts) == 1:
            return UserPath(*parts)
        elif len(parts) == 2:
            return ContentPath(*parts)
        elif len(parts) == 3:
            return BundlePath(*parts)
        elif len(parts) == 4:
            return BundleFilePath(*parts)
        else:
            raise ValueError(f"Unable to parse path: {path}")

    def _get_entity_from_path(self, path):
        parsed = self.parse_path(path)

        # guard against empty paths
        if isinstance(parsed, EmptyPath):
            raise ValueError(f"Cannot fetch root path: {path}")

        # note this sequence of ifs is essentially a case statement going down
        # a line from parent -> child -> grandchild
        if isinstance(parsed, UserPath):
            crnt = user = self._get_user_from_name(parsed.username)

        if isinstance(parsed, ContentPath):
            user_guid = user["guid"]

            # user_guid + content name should uniquely identify content, but
            # double check to be safe.
            crnt = content = self._get_content_from_name(user_guid, parsed.content)

        if isinstance(parsed, BundlePath):
            content_guid = content["guid"]
            crnt = self._get_content_bundle(content_guid, parsed.bundle)

        return crnt

    def _get_content_from_name(self, user_guid, content_name):
        """Fetch a single content entity."""

        # user_guid + content name should uniquely identify content, but
        # double check to be safe.
        contents = self.api.get_content(user_guid, content_name)
        if len(contents) != 1:
            err = (
                RsConnectApiMissingContentError
                if len(contents) == 0
                else RsConnectApiResultError
            )
            raise err(
                f"Expecting 1 content entry, but found {len(contents)}: {contents}"
            )

        res = contents[0]

        self._content_name_cache[(user_guid, content_name)] = res["guid"]
        return res

    def _get_content_bundle(self, content_guid, bundle_id):
        """Fetch a content bundle."""

        try:
            bundle = self.api.get_content_bundle(content_guid, bundle_id)
        except RsConnectApiRequestError as e:
            if e.args[0]["code"] == RSC_CODE_OBJECT_DOES_NOT_EXIST:
                raise RsConnectApiMissingContentError(
                    f"No bundle {bundle_id} for content {content_guid}"
                )
            raise e

        return bundle

    def _get_user_from_name(self, name):
        """Fetch a single user entity from user name."""
        users = self.api.get_users(prefix=name)
        try:
            user = next(iter([x for x in users if x["username"] == name]))

            self._user_name_cache[user["username"]] = user["guid"]

            return user
        except StopIteration:
            raise ValueError(f"No user named {name} found.")



import sys

from pins import board_folder
from pins import data

path_py, path_r = sys.argv[1], sys.argv[2]

# create board ----

board_py = board_folder(path_py)
board_r = board_folder(path_r)


# check pins ----

print("Checking mtcars pin")

res_mtcars = board_r.pin_read("mtcars")
assert res_mtcars.equals(data.mtcars)

meta_mtcars_py = board_py.pin_meta("mtcars")
print("\nPython meta:\n")
print(meta_mtcars_py)

meta_mtcars_r = board_r.pin_meta("mtcars")
print("\nR meta:\n")
print(meta_mtcars_r)



import pandas as pd

from pathlib import Path
from pins.rsconnect.fs import PinBundleManifest
from pins.meta import MetaFactory

p_root = Path("pins/tests/example-bundle")
p_root.parent.mkdir(parents=True, exist_ok=True)

p_index = p_root / "index.html"
p_index.write_text("<html><body>yo</body></html>")

p_data = p_root / "data_frame.csv"
df = pd.DataFrame({"x": [1, 2, 3]})
df.to_csv(p_data)

p_meta = p_root / "data.txt"
meta = MetaFactory().create(
    str(p_data), "csv", title="some title", name="data_frame.csv"
)
meta.to_yaml(p_meta.open("w"))

# add manifest last, since it enumerates all the files
# this lets you download them individually from rsconnect
PinBundleManifest.add_manifest_to_directory(str(p_root))



import os
import pandas as pd
import pytest

from pandas.testing import assert_frame_equal
from pathlib import Path

from pins import constructors as c
from pins.tests.conftest import (
    PATH_TO_EXAMPLE_BOARD,
    PATH_TO_EXAMPLE_VERSION,
    EXAMPLE_REL_PATH,
)
from pins.tests.helpers import rm_env


@pytest.fixture
def df_csv():
    return pd.read_csv(PATH_TO_EXAMPLE_VERSION / "df_csv.csv")


def check_dir_writable(p_dir):
    assert p_dir.parent.exists()
    assert os.access(p_dir.parent.absolute(), os.W_OK)


def check_cache_file_path(p_file, p_cache):
    rel_path = p_file.relative_to(p_cache)

    # parents has every entry you'd get if you called .parents all the way to some root.
    # for a relative path, the root is likely ".", so we subtract 1 to get the number
    # of parent directories.
    # note this essentially counts slashes, in a inter-OS friendly way.
    n_parents = len(rel_path.parents) - 1
    assert n_parents == 2


def construct_from_board(board):
    prot = board.fs.protocol
    fs_name = prot if isinstance(prot, str) else prot[0]

    if fs_name in ["file", ("file", "local")]:
        board = c.board_folder(board.board)
    elif fs_name == "rsc":
        board = c.board_rsconnect(
            server_url=board.fs.api.server_url, api_key=board.fs.api.api_key
        )
    elif fs_name == "abfs":
        board = c.board_azure(board.board)
    else:
        board = getattr(c, f"board_{fs_name}")(board.board)

    return board


# End-to-end constructor tests

# there are two facets of boards: reading and writing.
# copied from test_compat
@pytest.mark.skip_on_github
def test_constructor_board_url_data(tmp_cache, http_example_board_path, df_csv):
    board = c.board_url(
        http_example_board_path,
        # could derive from example version path
        pin_paths={"df_csv": "df_csv/20220214T163720Z-9bfad/"},
    )

    df = board.pin_read("df_csv")

    # check data ----
    assert_frame_equal(df, df_csv)


@pytest.mark.xfail
@pytest.mark.skip_on_github
def test_constructor_board_url_cache(
    tmp_cache, http_example_board_path, df_csv, tmp_path
):
    # TODO: downloading a pin does not put files in the same directory, since
    # in this case we are hashing on the full url.

    board = c.board_url(
        http_example_board_path,
        # could derive from example version path
        pin_paths={"df_csv": "df_csv/20220214T163718Z-eceac/"},
    )

    board.pin_read("df_csv")

    # cannot write or view pin versions

    with pytest.raises(NotImplementedError):
        board.pin_write(df_csv)
    with pytest.raises(NotImplementedError):
        board.pin_versions("df_csv")
    with pytest.raises(NotImplementedError):
        board.pin_version_delete(name="df_csv", version="20220214T163718Z")
    with pytest.raises(NotImplementedError):
        df = pd.DataFrame({"x": [1, 2, 3]})
        path = tmp_path / "data.csv"
        df.to_csv(path, index=False)
        board.pin_upload(path, "cool_pin")

    # check cache ----
    http_dirs = list(tmp_cache.glob("http_*"))

    assert len(http_dirs) == 1

    # there are two files in the flat cache (metadata, and the csv)
    parent = http_dirs[0]
    res = list(parent.rglob("*"))
    assert len(res) == 2

    # validate that it creates an empty metadata file
    assert len(x for x in res if x.endswith("df_csv.csv")) == 1
    assert len(x for x in res if x.endswith("data.txt")) == 1

    assert len(list(parent.glob("**/*"))) == 2


@pytest.mark.skip_on_github
def test_constructor_board_url_file(tmp_cache, http_example_board_path):
    # TODO: downloading a pin does not put files in the same directory, since
    # in this case we are hashing on the full url.

    board = c.board_url(
        http_example_board_path,
        # could derive from example version path
        pin_paths={"df_csv": "df_csv/20220214T163718Z-eceac/df_csv.csv"},
    )

    board.pin_download("df_csv")

    # check cache ----
    http_dirs = list(tmp_cache.glob("http_*"))

    assert len(http_dirs) == 1

    # there are two files in the flat cache (metadata, and the csv)
    parent = http_dirs[0]
    res = list(parent.rglob("*"))
    assert len(res) == 1

    assert str(res[0]).endswith("df_csv.csv")

    new_board = eval(c.board_deparse(board), c.__dict__)
    assert new_board.pin_list() == board.pin_list()


@pytest.mark.skip_on_github
def test_constructor_board_github(tmp_cache, http_example_board_path, df_csv):
    board = c.board_github("rstudio", "pins-python", EXAMPLE_REL_PATH)  # noqa

    df = board.pin_read("df_csv")
    assert_frame_equal(df, df_csv)

    cache_options = list(tmp_cache.glob("github_*"))
    assert len(cache_options) == 1
    cache_dir = cache_options[0]

    res = list(cache_dir.rglob("**/*.csv"))
    assert len(res) == 1

    check_cache_file_path(res[0], cache_dir)


@pytest.fixture(scope="function")
def board(backend):
    # TODO: copied from test_compat.py

    board = backend.create_tmp_board(str(PATH_TO_EXAMPLE_BOARD.absolute()))
    yield board
    backend.teardown_board(board)


def test_constructor_boards(board, df_csv, tmp_cache):
    # TODO: would be nice to have fixtures for each board constructor
    # doesn't need to copy over pins-compat content

    # create board from constructor -------------------------------------------
    board = construct_from_board(board)

    # read a pin and check its contents ---------------------------------------

    df = board.pin_read("df_csv")

    # check data
    assert_frame_equal(df, df_csv)

    # check the cache structure -----------------------------------------------

    # check cache
    if board.fs.protocol in ["file", ("file", "local")]:
        # no caching for local file boards
        pass
    else:
        # check path structure ----

        options = list(tmp_cache.glob("*"))
        assert len(options) == 1

        cache_dir = options[0]
        res = list(cache_dir.rglob("*/*.csv"))
        assert len(res) == 1

        check_cache_file_path(res[0], cache_dir)

        # check cache touch on access time ----

        meta = board.pin_meta("df_csv")
        p_cache_meta = (
            Path(board._get_cache_path(meta.name, meta.version.version)) / "data.txt"
        )
        orig_access = p_cache_meta.stat().st_atime

        board.pin_meta("df_csv")

        new_access = p_cache_meta.stat().st_atime

        assert orig_access < new_access


@pytest.fixture(scope="function")
def board2(backend):
    board2 = backend.create_tmp_board()
    yield board2
    backend.teardown_board(board2)


def test_constructor_boards_multi_user(board2, df_csv, tmp_cache):
    prot = board2.fs.protocol
    fs_name = prot if isinstance(prot, str) else prot[0]

    if fs_name == "rsc":
        # TODO: RSConnect writes pin names like <user>/<name>, so would need to
        # modify test
        pytest.skip()
    elif fs_name == "abfs":
        fs_name = "azure"

    first = construct_from_board(board2)

    first.pin_write(df_csv, "df_csv", type="csv")
    assert first.pin_list() == ["df_csv"]

    second = construct_from_board(board2)
    second.pin_write(df_csv, "another_df_csv", type="csv")

    assert sorted(second.pin_list()) == sorted(["df_csv", "another_df_csv"])


# Board particulars ===========================================================


@pytest.mark.skip_on_github
def test_board_constructor_local_default_writable():

    with rm_env("PINS_DATA_DIR"):
        board = c.board_local()
        p_board = Path(board.board)

        check_dir_writable(p_board)
        assert p_board.name == "pins-py"


def test_board_constructor_temp_writable():
    with rm_env("PINS_DATA_DIR"):
        board = c.board_temp()
        p_board = Path(board.board)

        check_dir_writable(p_board)
        assert len(list(p_board.glob("*"))) == 0


def test_board_constructor_folder(tmp_dir2, df):
    board = c.board_folder(str(tmp_dir2))
    board.pin_write(df, "some_df", type="csv")

    assert (tmp_dir2 / "some_df").exists()
    df2 = board.pin_read("some_df")

    assert df.equals(df2)


# Deparsing ===================================================================


def test_board_deparse(board):
    prot = board.fs.protocol

    with rm_env("CONNECT_API_KEY"):
        if prot == "rsc":
            os.environ["CONNECT_API_KEY"] = board.fs.api.api_key

        new_board = eval(c.board_deparse(board), c.__dict__)
        new_board.pin_list()



from typing import ClassVar
from dataclasses import dataclass, asdict, field, fields, InitVar
from pathlib import Path

import yaml

from typing import Mapping, Union, Sequence, Optional, List

from .versions import VersionRaw, Version, guess_version
from ._types import StrOrFile, IOBase

META_FILENAME = "data.txt"
DEFAULT_API_VERSION = 1


@dataclass
class MetaRaw:
    """Absolute minimum metadata for a pin.

    Parameters
    ----------
    file:
        All relevant files contained in the pin. Note that these be absolute paths
        to fetch from the target filesystem.
    type:
        The type of pin data stored. This is used to determine how to read / write it.
    """

    file: "str | Sequence[str] | None"
    type: str
    name: str


@dataclass
class Meta:
    """Represent metadata for a pin version.

    Parameters
    ----------
    title:
        A title for the pin.
    description:
        A detailed description of the pin contents.
    tags:
        Optional tags applied to the pin.
    created:
        Datetime the pin was created (TODO: document format).
    pin_hash:
        A hash of the pin.
    file:
        All relevant files in the pin. Should be relative to this pin's folder.
    file_size:
        The total size of the files in the pin.
    type:
        The type of pin data stored. This is used to determine how to read / write it.
    api_version:
        The internal version of the metadata format.
    name:
        TODO - where is this in R pins?
    user:
        A dictionary of additional metadata that may be specified by the user.
    local:
        A dictionary of additional metadata that may be added by the board, depending
        on the backend used. E.g. RStudio Connect content id, url, etc..

    """

    _excluded: ClassVar["set[str]"] = {"name", "version", "local"}

    title: Optional[str]
    description: Optional[str]

    # TODO(defer): different from R pins, which has a local field
    created: str
    pin_hash: str

    file: Union[str, Sequence[str]]
    file_size: int
    type: str

    api_version: int

    # In the metadata yaml, the created field uses a custom format, so
    # we need a version object in order to render it. You can think of
    # the version here as "the thing that was used to create version_name,
    # pin_hash, created, etc.."
    version: VersionRaw

    tags: Optional[List[str]] = None
    name: Optional[str] = None
    user: Mapping = field(default_factory=dict)
    local: Mapping = field(default_factory=dict)

    unknown_fields: InitVar["dict | None"] = None

    def __post_init__(self, unknown_fields: "dict | None"):
        unknown_fields = {} if unknown_fields is None else unknown_fields

        self._unknown_fields = unknown_fields

    def __getattr__(self, k):
        try:
            return self._unknown_fields[k]
        except KeyError:
            raise AttributeError(f"No metadata field not found: {k}")

    def to_dict(self) -> Mapping:
        data = asdict(self)

        return data

    def to_pin_dict(self):
        d = self.to_dict()

        for k in self._excluded:
            del d[k]

        # TODO: once tag writing is implemented, delete this line
        del d["tags"]

        return d

    @classmethod
    def from_pin_dict(cls, data, pin_name, version, local=None) -> "Meta":
        # TODO: re-arrange Meta argument positions to reflect what's been
        # learned about default arguments. e.g. title was not used at some
        # point in api_version 1
        all_field_names = {entry.name for entry in fields(Meta)}

        keep_fields = all_field_names - cls._excluded

        extra = {"title": None} if "title" not in data else {}
        local = {} if local is None else local

        meta_data = {k: v for k, v in data.items() if k in keep_fields}
        unknown = {k: v for k, v in data.items() if k not in keep_fields}

        return cls(
            **meta_data,
            **extra,
            name=pin_name,
            version=version,
            local=local,
            unknown_fields=unknown,
        )

    def to_pin_yaml(self, f: Optional[IOBase] = None) -> "str | None":
        data = self.to_pin_dict()

        return yaml.dump(data, f)


@dataclass
class MetaV0:
    file: Union[str, Sequence[str]]
    type: str
    description: "str | None"

    name: str

    version: VersionRaw

    # holds raw data.txt contents
    original_fields: dict = field(default_factory=dict)
    user: dict = field(default_factory=dict, init=False)
    local: Mapping = field(default_factory=dict)

    title: ClassVar[None] = None
    created: ClassVar[None] = None
    pin_hash: ClassVar[None] = None
    file_size: ClassVar[None] = None
    api_version: ClassVar[None] = None

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_pin_dict(cls, data, pin_name, version, local=None) -> "MetaV0":
        # could infer from dataclasses.fields(), but seems excessive.
        req_fields = {"type", "description"}

        # Note that we need to .get(), since fields may not be in metadata
        req_inputs = {k: data.get(k) for k in req_fields}
        req_inputs["file"] = data["path"]

        local = {} if local is None else local
        return cls(
            **req_inputs,
            name=pin_name,
            original_fields=data,
            version=version,
            local=local,
        )

    def to_pin_dict(self):
        raise NotImplementedError("v0 pins metadata are read only.")

    def to_pin_yaml(self, *args, **kwargs):
        self.to_pin_dict()


class MetaFactory:
    """Responsible for creating and loading (e.g. from yaml) of meta objects."""

    def get_meta_name(self, *args, **kwargs) -> str:
        return META_FILENAME

    def get_version_for_meta(self, api_version) -> Version:
        if api_version != 1:
            raise NotImplementedError("Unsupported api_version: %s" % api_version)

        return Version

    def create(
        self,
        base_folder: "str | Path",
        files: Sequence[StrOrFile],
        type,
        # TODO: when files is a string name should be okay as None
        name,
        title,
        description=None,
        created=None,
        user=None,
    ) -> Meta:

        if title is None:
            raise NotImplementedError("title arguments required")
        if isinstance(files, str):
            from pathlib import Path

            version = Version.from_files([files], created)
            p_file = Path(files)
            file_size = p_file.stat().st_size
            file_name = str(Path(files).relative_to(Path(base_folder)))

        elif isinstance(files, IOBase):
            # TODO: in theory can calculate size from a file object, but let's
            # wait until it's clear how calculating file size fits into pins
            # e.g. in combination with folders, etc..

            # from os import fstat
            #
            # version = Version.from_files([files], created)
            # files_size = fstat(files.fileno()).st_size

            raise NotImplementedError("Cannot create from file object.")
        else:
            raise NotImplementedError("TODO: creating meta from multiple files")

        return Meta(
            title=title,
            description=description,
            file=file_name,  # TODO: FINISH
            file_size=file_size,
            pin_hash=version.hash,
            created=version.render_created(),
            type=type,
            api_version=DEFAULT_API_VERSION,
            name=name,
            user=user if user is not None else {},
            version=version,
        )

    def create_raw(self, files: Sequence[StrOrFile], type: str, name: str) -> MetaRaw:
        return MetaRaw(files, type, name)

    def read_pin_yaml(
        self,
        f: IOBase,
        pin_name: str,
        version: "str | VersionRaw",
        local=None,
    ) -> Meta:
        if isinstance(version, str):
            version_obj = guess_version(version)
        else:
            version_obj = version

        data = yaml.safe_load(f)

        api_version = data.get("api_version", 0)
        if api_version >= 2:
            raise NotImplementedError(
                f"api_version {api_version} by this version of the pins library"
            )
        elif api_version == 0:
            cls_meta = MetaV0
        else:
            cls_meta = Meta

        return cls_meta.from_pin_dict(data, pin_name, version=version_obj, local=local)



import time

import pytest
from pins.cache import (
    CachePruner,
    touch_access_time,
    cache_prune,
    PinsCache,
    PinsUrlCache,
)

from fsspec import filesystem
from pathlib import Path

# NOTE: windows time.time() implementation appears to have 16 millisecond precision, so
# we need to add a small delay, in order to avoid prune checks appearing to happen at the
# exact same moment something earlier was created / accessed.
# see: https://stackoverflow.com/a/1938096/1144523


# Utilities ===================================================================


def _sleep():
    # time-based issues keep arising erratically in windows checks, so try to shoot
    # well past
    time.sleep(0.3)


@pytest.fixture
def some_file(tmp_dir2):
    p = tmp_dir2 / "some_file.txt"
    p.touch()
    return p


def test_touch_access_time_manual(some_file):
    some_file.stat().st_atime

    access_time = time.time() - 60 * 60 * 24
    touch_access_time(some_file, access_time)

    assert some_file.stat().st_atime == access_time


def test_touch_access_time_auto(some_file):
    orig_access = some_file.stat().st_atime

    _sleep()
    new_time = touch_access_time(some_file)

    assert some_file.stat().st_atime == new_time
    assert orig_access < new_time


# Cache Classes ===============================================================

# Boards w/ default cache =====================================================


def test_pins_cache_hash_name_preserves():
    cache = PinsCache(fs=filesystem("file"), hash_prefix="")
    assert cache.hash_name("a/b/c.txt") == Path("a/b/c.txt")


def test_pins_cache_url_hash_name():
    cache = PinsUrlCache(fs=filesystem("file"))
    hashed = cache.hash_name("http://example.com/a.txt", True)

    p_hash = Path(hashed)

    # should have form <url_hash>/<version_placeholder>/<filename>
    assert p_hash.name == "a.txt"

    # count parent dirs, excluding root (e.g. "." or "/")
    n_parents = len(p_hash.parents) - 1
    assert n_parents == 2


@pytest.mark.skip("TODO")
def test_pins_cache_open():
    # check that opening works and creates the cached file
    pass


# Cache pruning ===============================================================


@pytest.fixture
def a_cache(tmp_dir2):
    return tmp_dir2 / "board_cache"


def create_metadata(p, access_time):
    p.mkdir(parents=True, exist_ok=True)
    meta = p / "data.txt"
    meta.touch()
    touch_access_time(meta, access_time)


@pytest.fixture
def pin1_v1(a_cache):  # current
    v1 = a_cache / "a_pin" / "version_1"
    create_metadata(v1, time.time())

    return v1


@pytest.fixture
def pin1_v2(a_cache):
    v2 = a_cache / "a_pin" / "version_2"
    create_metadata(v2, time.time() - 60 * 60 * 24)  # one day ago

    return v2


@pytest.fixture
def pin2_v3(a_cache):
    v3 = a_cache / "other_pin" / "version_3"
    create_metadata(v3, time.time() - 60 * 60 * 48)  # two days ago

    return v3


def test_cache_pruner_old_versions_none(a_cache, pin1_v1):
    _sleep()

    pruner = CachePruner(a_cache)

    old = pruner.old_versions(days=1)

    assert len(old) == 0


def test_cache_pruner_old_versions_days0(a_cache, pin1_v1):
    _sleep()

    pruner = CachePruner(a_cache)
    old = pruner.old_versions(days=0)

    assert len(old) == 1
    assert old[0] == pin1_v1


def test_cache_pruner_old_versions_some(a_cache, pin1_v1, pin1_v2):
    _sleep()

    # create: tmp_dir/pin1/version1

    pruner = CachePruner(a_cache)

    old = pruner.old_versions(days=1)

    assert len(old) == 1
    assert old[0] == pin1_v2


def test_cache_pruner_old_versions_multi_pins(a_cache, pin1_v2, pin2_v3):
    _sleep()

    pruner = CachePruner(a_cache)
    old = pruner.old_versions(days=1)

    assert len(old) == 2
    assert set(old) == {pin1_v2, pin2_v3}


def test_cache_prune_prompt(a_cache, pin1_v1, pin2_v3, monkeypatch):
    _sleep()

    cache_prune(days=1, cache_root=a_cache.parent, prompt=False)

    versions = list(a_cache.glob("*/*"))

    # pin2_v3 deleted
    assert len(versions) == 1



import fsspec
import os
import uuid
import pandas as pd
import pytest

from pins.tests.helpers import DEFAULT_CREATION_DATE, rm_env
from pins.config import PINS_ENV_INSECURE_READ
from pins.errors import PinsError, PinsInsecureReadError
from pins.meta import MetaRaw

from datetime import datetime, timedelta
from time import sleep
from pathlib import Path

# using pytest cases, so that we can pass in fixtures as parameters
# TODO: this seems like maybe overkill
from pytest_cases import fixture, parametrize


@fixture
def df():
    import pandas as pd

    return pd.DataFrame({"x": [1, 2, 3], "y": ["a", "b", "c"]})


@fixture
def board(backend):
    yield backend.create_tmp_board()
    backend.teardown()


@fixture
def board_with_cache(backend):
    from pins.constructors import board as board_constructor, board_rsconnect

    board = backend.create_tmp_board()

    if backend.fs_name == "rsc":
        # The rsconnect board is special, in that it's slower to set up and tear down,
        # so our test suite uses multiple rsconnect users in testing its API, and
        # board behavior. As a result, we need to pass the credentials directly in.
        server_url, api_key = board.fs.api.server_url, board.fs.api.api_key
        board_with_cache = board_rsconnect(server_url=server_url, api_key=api_key)
    else:
        board_with_cache = board_constructor(backend.fs_name, board.board)

    yield board_with_cache

    backend.teardown()


# misc ========================================================================


def test_board_validate_pin_name_root(board):
    with pytest.raises(ValueError) as exc_info:
        board.path_to_pin("/some_pin")

    assert "Invalid pin name" in exc_info.value.args[0]


# pin_write ===================================================================


def test_board_pin_write_default_title(board):

    df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})
    meta = board.pin_write(df, "df_csv", title=None, type="csv")
    assert meta.title == "df_csv: a pinned 3 x 2 DataFrame"


def test_board_pin_write_prepare_pin(board, tmp_dir2):

    df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})

    meta = board.prepare_pin_version(
        str(tmp_dir2), df, "df_csv", title=None, type="csv"
    )
    assert meta.file == "df_csv.csv"
    assert (tmp_dir2 / "data.txt").exists()
    assert (tmp_dir2 / "df_csv.csv").exists()
    assert not (tmp_dir2 / "df_csv.csv").is_dir()


def test_board_pin_write_roundtrip(board):

    df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})

    assert not board.pin_exists("df_csv")

    board.pin_write(df, "df_csv", type="csv")

    assert board.pin_exists("df_csv")

    loaded_df = board.pin_read("df_csv")
    assert loaded_df.equals(df)


def test_board_pin_write_type_not_specified_error(board):
    class C:
        pass

    with pytest.raises(NotImplementedError):
        board.pin_write(C(), "cool_pin")


def test_board_pin_write_type_error(board):
    class C:
        pass

    with pytest.raises(NotImplementedError) as exc_info:
        board.pin_write(C(), "cool_pin", type="MY_TYPE")

    assert "MY_TYPE" in exc_info.value.args[0]


def test_board_pin_write_feather_deprecated(board):
    df = pd.DataFrame({"x": [1, 2, 3]})

    with pytest.warns(DeprecationWarning):
        board.pin_write(df, "cool_pin", type="feather")


def test_board_pin_write_file_raises_error(board, tmp_path):
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path.joinpath("data.csv")
    df.to_csv(path, index=False)

    # TODO: should this error?
    with pytest.raises(NotImplementedError):
        board.pin_write(path, "cool_pin", type="file")


def test_board_pin_download(board_with_cache, tmp_path):
    # create and save data
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path / "data.csv"
    df.to_csv(path, index=False)

    meta = board_with_cache.pin_upload(path, "cool_pin")
    assert meta.type == "file"

    (pin_path,) = board_with_cache.pin_download("cool_pin")
    df = pd.read_csv(pin_path)
    assert df.x.tolist() == [1, 2, 3]

    with pytest.raises(NotImplementedError):
        board_with_cache.pin_read("cool_pin")


def test_board_pin_download_filename_many_suffixes(board_with_cache, tmp_path):
    # create and save data
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path / "data.a.b.csv"
    df.to_csv(path, index=False)

    board_with_cache.pin_upload(path, "cool_pin")

    (pin_path,) = board_with_cache.pin_download("cool_pin")
    assert Path(pin_path).name == "data.a.b.csv"

    df = pd.read_csv(pin_path)
    assert df.x.tolist() == [1, 2, 3]


def test_board_pin_download_filename_no_suffixes(board_with_cache, tmp_path):
    # create and save data
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path / "data"
    df.to_csv(path, index=False)

    board_with_cache.pin_upload(path, "cool_pin")

    (pin_path,) = board_with_cache.pin_download("cool_pin")
    assert Path(pin_path).name == "data"

    df = pd.read_csv(pin_path)
    assert df.x.tolist() == [1, 2, 3]


def test_board_pin_download_filename(board_with_cache, tmp_path):
    # create and save data
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path / "data.csv"
    df.to_csv(path, index=False)

    meta = board_with_cache.pin_upload(path, "cool_pin")

    assert meta.file == "data.csv"

    (pin_path,) = board_with_cache.pin_download("cool_pin")
    assert Path(pin_path).name == "data.csv"


def test_board_pin_download_no_cache_error(board, tmp_path):
    df = pd.DataFrame({"x": [1, 2, 3]})
    path = tmp_path / "data.csv"
    df.to_csv(path, index=False)

    # TODO: should this error?
    meta = board.pin_upload(path, "cool_pin")
    assert meta.type == "file"

    # file boards work okay, since the board directory itself is the cache
    if board.fs.protocol in ["file", ("file", "local")]:
        pytest.skip()

    # uncached boards should fail, since nowhere to store the download
    with pytest.raises(PinsError):
        (pin_path,) = board.pin_download("cool_pin")


def test_board_pin_upload_path_list(board_with_cache, tmp_path):
    # create and save data
    df = pd.DataFrame({"x": [1, 2, 3]})

    path = tmp_path / "data.csv"
    df.to_csv(path, index=False)

    meta = board_with_cache.pin_upload([path], "cool_pin")
    assert meta.type == "file"

    (pin_path,) = board_with_cache.pin_download("cool_pin")


def test_board_pin_write_rsc_index_html(board, tmp_dir2, snapshot):
    if board.fs.protocol != "rsc":
        pytest.skip()

    df = pd.DataFrame({"x": [1, 2, None], "y": ["a", "b", "c"]})

    pin_name = "test_rsc_pin"

    board.prepare_pin_version(
        str(tmp_dir2),
        df,
        pin_name,
        type="csv",
        title="some pin",
        description="some description",
        created=DEFAULT_CREATION_DATE,
    )

    snapshot.assert_equal_dir(tmp_dir2)


# pin_write against different types -------------------------------------------


@parametrize(
    "obj, type_",
    [
        (df, "csv"),
        (df, "joblib"),
        ({"a": 1, "b": [2, 3]}, "joblib"),
        ({"a": 1, "b": [2, 3]}, "json"),
    ],
)
def test_board_pin_write_type(board, obj, type_, request):
    with rm_env(PINS_ENV_INSECURE_READ):
        os.environ[PINS_ENV_INSECURE_READ] = "1"
        meta = board.pin_write(obj, "test_pin", type=type_, title="some title")
        dst_obj = board.pin_read("test_pin")

        assert meta.type == type_

        if isinstance(obj, pd.DataFrame):
            assert obj.equals(dst_obj)

        obj == dst_obj


def test_board_pin_read_insecure_fail_default(board):
    board.pin_write({"a": 1}, "test_pin", type="joblib", title="some title")
    with pytest.raises(PinsInsecureReadError) as exc_info:
        board.pin_read("test_pin")

    assert "joblib" in exc_info.value.args[0]


def test_board_pin_read_insecure_fail_board_flag(board):
    # board flag prioritized over env var
    with rm_env(PINS_ENV_INSECURE_READ):
        os.environ[PINS_ENV_INSECURE_READ] = "1"
        board.allow_pickle_read = False
        board.pin_write({"a": 1}, "test_pin", type="joblib", title="some title")
        with pytest.raises(PinsInsecureReadError):
            board.pin_read("test_pin")


def test_board_pin_read_insecure_succeed_board_flag(board):
    # board flag prioritized over env var
    with rm_env(PINS_ENV_INSECURE_READ):
        os.environ[PINS_ENV_INSECURE_READ] = "0"
        board.allow_pickle_read = True
        board.pin_write({"a": 1}, "test_pin", type="joblib", title="some title")
        board.pin_read("test_pin")


# pin_delete ==================================================================


@pytest.fixture
def pin_name():
    return str(uuid.uuid4())


@pytest.fixture
def pin_del(board, df, pin_name):
    meta_old = board.pin_write(df, pin_name, type="csv", title="some title")
    sleep(1)
    meta_new = board.pin_write(df, pin_name, type="csv", title="some title")

    assert len(board.pin_versions(pin_name)) == 2
    assert meta_old.version.version != meta_new.version.version

    return meta_old, meta_new


@pytest.fixture
def pin_prune(board, df, pin_name):
    today = datetime.now()
    day_ago = today - timedelta(days=1, minutes=1)
    two_days_ago = today - timedelta(days=2, minutes=1)

    board.pin_write(df, pin_name, type="csv", title="some title", created=today)
    board.pin_write(df, pin_name, type="csv", title="some title", created=day_ago)
    board.pin_write(df, pin_name, type="csv", title="some title", created=two_days_ago)

    versions = board.pin_versions(pin_name, as_df=False)
    assert len(versions) == 3

    return versions


def test_board_pin_delete(board, df, pin_name, pin_del):
    board.pin_delete(pin_name)

    assert board.pin_exists(pin_name) is False


def test_board_pin_version_delete_older(board, pin_name, pin_del):
    meta_old, meta_new = pin_del

    board.pin_version_delete(pin_name, meta_old.version.version)
    df_versions = board.pin_versions(pin_name)

    # Note that using `in` on a pandas Series checks against the index :/
    assert meta_old.version.version not in df_versions.version.values
    assert meta_new.version.version in df_versions.version.values


def test_board_pin_version_delete_latest(board, pin_name, pin_del):
    meta_old, meta_new = pin_del

    if board.fs.protocol == "rsc":
        with pytest.raises(PinsError) as exc_info:
            board.pin_version_delete(pin_name, meta_new.version.version)

            "cannot delete the latest pin version" in exc_info.value.args[0]
        return

    board.pin_version_delete(pin_name, meta_new.version.version)
    df_versions = board.pin_versions(pin_name)

    # Note that using `in` on a pandas Series checks against the index :/
    assert meta_old.version.version in df_versions.version.values
    assert meta_new.version.version not in df_versions.version.values


@pytest.mark.parametrize("n", [1, 2])
def test_board_pin_versions_prune_n(board, pin_prune, pin_name, n):

    board.pin_versions_prune(pin_name, n=n)
    new_versions = board.pin_versions(pin_name, as_df=False)

    assert len(new_versions) == n

    # TODO(compat): versions are currently reversed from R pins, with latest last
    # so we need to reverse to check the n latest versions
    rev_vers = list(reversed(pin_prune))
    for ii, v in enumerate(reversed(new_versions)):
        assert rev_vers[ii].version == v.version


@pytest.mark.parametrize("days", [1, 2])
def test_board_pin_versions_prune_days(board, pin_prune, pin_name, days):

    # RStudio cannot handle days, since it involves pulling metadata
    if board.fs.protocol == "rsc":
        with pytest.raises(NotImplementedError):
            board.pin_versions_prune(pin_name, days=days)
        return

    board.pin_versions_prune(pin_name, days=days)

    new_versions = board.pin_versions(pin_name, as_df=False)

    # each of the 3 versions adds an 1 more day + 1 min
    assert len(new_versions) == days


# pin_search ==================================================================


@pytest.mark.parametrize(
    "search, matches",
    [
        # beginning character
        ("x", ["x-pin-1", "x-pin-2"]),
        # middle of name
        ("pin", ["x-pin-1", "x-pin-2", "y-pin-1"]),
        # regex set
        ("[0-9]", ["x-pin-1", "x-pin-2", "y-pin-1"]),
        # exists only in title
        ("the-title", ["x-pin-1", "x-pin-2", "y-pin-1", "y-z"]),
    ],
)
def test_board_pin_search_name(board, df, search, matches):
    if board.fs.protocol == "rsc":
        matches = ["derek/" + m for m in matches]

        # rsc doesn't search by title
        if search in ["the-title", "[0-9]"]:
            pytest.xfail()

    for name in ["x-pin-1", "x-pin-2", "y-pin-1", "y-z"]:
        board.pin_write(df, name, type="csv", title="the-title")

    metas = board.pin_search(search, as_df=False)
    sorted_meta_names = sorted([m.name for m in metas])
    assert sorted_meta_names == sorted(matches)


# BaseBoard specific ==========================================================

from pins.boards import BaseBoard  # noqa
from pins.cache import PinsCache  # noqa


def test_board_base_pin_meta_cache_touch(tmp_dir2, df):

    cache = fsspec.filesystem(
        "pinscache",
        target_protocol="file",
        same_names=True,
        hash_prefix=str(tmp_dir2),
    )
    board = BaseBoard(str(tmp_dir2), fs=cache)

    board.pin_write(df, "some_df", type="csv")
    meta = board.pin_meta("some_df")
    v = meta.version.version

    p_cache_version = board._get_cache_path(meta.name, v)
    p_cache_meta = Path(p_cache_version) / "data.txt"

    orig_access = p_cache_meta.stat().st_atime

    sleep(0.3)

    board.pin_meta("some_df")

    new_access = p_cache_meta.stat().st_atime

    assert orig_access < new_access


# RStudio Connect specific ====================================================

# import fixture that builds / tearsdown user "susan"
from pins.tests.test_rsconnect_api import (  # noqa
    fs_short,
    fs_admin,
    rsc_admin,
    rsc_short,
)
from pins.boards import BoardRsConnect  # noqa


@pytest.mark.fs_rsc
@pytest.fixture
def board_short(fs_short):  # noqa
    board_short = BoardRsConnect("", fs_short)
    return board_short


@pytest.mark.fs_rsc
def test_board_pin_write_rsc_full_name(df, board_short):  # noqa
    board_short.pin_write(df, "susan/some_df", type="csv")


@pytest.mark.fs_rsc
def test_board_pin_search_admin_user(df, board_short, fs_admin):  # noqa
    board_short.pin_write(df, "some_df", type="csv")

    board_admin = BoardRsConnect("", fs_admin)
    search_res = board_admin.pin_search("susan", as_df=False)

    assert len(search_res) == 1
    assert search_res[0].name == "susan/some_df"
    assert isinstance(search_res[0], MetaRaw)

    search_res2 = board_admin.pin_search("susan", as_df=True)
    assert search_res2.shape == (1, 6)
    assert search_res2.loc[0, "name"] == "susan/some_df"
    assert isinstance(search_res2.loc[0, "meta"], MetaRaw)


@pytest.mark.fs_rsc
def test_board_rsc_pin_write_title_update(df, board_short):
    board_short.pin_write(df, "susan/some_df", type="csv", title="title a")
    board_short.pin_write(df, "susan/some_df", type="csv", title="title b")

    content = board_short.fs.info("susan/some_df")
    assert content["title"] == "title b"


@pytest.mark.fs_rsc
def test_board_pin_meta_is_full_name(df, board_short):
    meta = board_short.pin_write(df, "susan/some_df", type="csv")

    assert meta.name == "susan/some_df"

    meta2 = board_short.pin_write(df, "some_df", type="csv")
    assert meta2.name == "susan/some_df"

    meta3 = board_short.pin_meta("some_df")
    assert meta3.name == "susan/some_df"


@pytest.mark.fs_rsc
def test_board_rsc_path_to_pin_safe(board_short):
    assert board_short.path_to_pin("me/some_pin") == "me/some_pin"


@pytest.mark.fs_rsc
def test_board_rsc_require_full_pin_name(board_short):
    # the tests set the special env var below to allow short pin names,
    # but here we test this is not what should happen by default.
    # we test that full names work with the RSC board above, so these two
    # things should cover us.
    from pins.config import PINS_ENV_ALLOW_RSC_SHORT_NAME

    with rm_env(PINS_ENV_ALLOW_RSC_SHORT_NAME):
        with pytest.raises(ValueError) as exc_info:
            board_short.validate_pin_name("mtcars")

        assert "Invalid pin name" in exc_info.value.args[0]

        with pytest.raises(ValueError) as exc_info:
            board_short.path_to_pin("mtcars")

        assert "Invalid pin name" in exc_info.value.args[0]


@pytest.mark.fs_rsc
def test_board_rsc_pin_write_other_user_fails(df, board_short):
    with pytest.raises(PinsError) as exc_info:
        board_short.pin_write(df, "derek/mtcarszzzzz")

    assert "a new piece of content for another user" in exc_info.value.args[0]


@pytest.mark.fs_rsc
def test_board_rsc_pin_write_acl(df, board_short):
    board_short.pin_write(df, "susan/mtcars", type="csv", access_type="all")
    content = board_short.fs.info("susan/mtcars")
    assert content["access_type"] == "all"


@pytest.mark.fs_rsc
def test_board_rsc_pin_read_public(df, board_short):
    from pins.boards import BoardManual

    board_short.pin_write(df, "susan/mtcars", type="csv", access_type="all")

    # note that users can also get this from the web ui
    content_url = board_short.fs.info("susan/mtcars")["content_url"]

    # shouldn't be a key set in env, but remove just in case
    fs = fsspec.filesystem("http")
    board_url = BoardManual("", fs, pin_paths={"rsc_public": content_url})

    df_no_key = board_url.pin_read("rsc_public")
    assert df_no_key.equals(df)


# Manual Board Specific =======================================================

from pins.boards import BoardManual  # noqa


def test_board_manual_http_file_download():

    path = "https://raw.githubusercontent.com/rstudio/pins-python"
    license_path = "main/LICENSE"

    # use a simple cache, which automatically creates a temporary directory
    fs = fsspec.filesystem(
        "simplecache", target_protocol="http", target_options={"block_size": 0}
    )

    # with path ----
    board = BoardManual(path, fs, pin_paths={"license": "main/LICENSE"})

    assert board.pin_list() == ["license"]
    # TODO: better assert
    assert len(board.pin_download("license"))

    # no base path ----
    board2 = BoardManual("", fs, pin_paths={"license": f"{path}/{license_path}"})

    assert board2.pin_list() == ["license"]
    # TODO better assert
    assert len(board2.pin_download("license"))


def test_board_manual_pin_read():
    # TODO: block size must be set to 0 to handle gzip encoding from github
    # see https://github.com/fsspec/filesystem_spec/issues/389
    fs = fsspec.filesystem("http", block_size=0)
    board = BoardManual(
        "https://raw.githubusercontent.com/rstudio/pins-python/main/pins/tests/pins-compat",
        fs,
        pin_paths={
            "df_csv": "df_csv/20220214T163718Z-eceac/",
            "df_csv2_v2": "df_csv/20220214T163720Z-9bfad/",
        },
    )

    df = board.pin_read("df_csv")

    # do a somewhat data-framey check
    assert df.shape[0] > 1


def test_board_manual_construct_path():
    fs = fsspec.filesystem("file")
    root = "pins/tests/pins-compat"
    path_df_csv = "df_csv/20220214T163718Z-eceac/"
    path_df_csv_v2 = "df_csv/20220214T163720Z-9bfad/df_csv.csv"

    board = BoardManual(
        root,
        fs,
        pin_paths={
            "df_csv": path_df_csv,
            "df_csv2_v2": path_df_csv_v2,
        },
    )

    # path to pin folder ----
    # creates path to pin, ignores version, can include data.txt
    assert board.construct_path(["df_csv"]) == f"{root}/{path_df_csv}"
    assert board.construct_path(["df_csv", "v"]) == f"{root}/{path_df_csv}"
    assert (
        board.construct_path(["df_csv", "v", "data.txt"])
        == f"{root}/{path_df_csv}data.txt"
    )

    with pytest.raises(NotImplementedError) as exc_info:
        board.construct_path(["df_csv", "v", "data.txt", "too_much"])

    assert "Unable to construct path" in exc_info.value.args[0]

    # path to individual file ----
    assert board.construct_path(["df_csv2_v2"]) == f"{root}/{path_df_csv_v2}"

    with pytest.raises(ValueError) as exc_info:
        board.construct_path(["df_csv2_v2", "v"])

    assert "assumed to be a single file" in exc_info.value.args[0]



import pytest
import tempfile
import yaml

from datetime import datetime
from io import StringIO

from pins.meta import Meta, MetaFactory
from pins.versions import Version

META_DEFAULTS = {
    "title": "some title",
    "description": "some description",
    "file": "some_file.csv",
    "file_size": 3,
    "pin_hash": "abcdef",
    "created": "20001230T124647Z",
    "type": "csv",
    "api_version": 1,
    "version": Version(datetime(2000, 12, 30, 12, 46, 47), "abcdef"),
}


@pytest.fixture
def meta():
    return Meta(**META_DEFAULTS)


@pytest.mark.xfail
def test_meta_to_dict_is_recursive(meta):
    d_meta = meta.to_dict()
    assert d_meta["version"] == meta.version.to_dict()


def test_meta_to_pin_dict_roundtrip(meta):
    d_meta = meta.to_pin_dict()
    meta2 = Meta.from_pin_dict(d_meta, meta.name, meta.version)
    assert meta == meta2


def test_meta_unknown_fields():
    m = Meta(**META_DEFAULTS, unknown_fields={"some_other_field": 1})

    assert m.some_other_field == 1

    with pytest.raises(AttributeError):
        m.should_not_exist_here

    assert "unknown_fields" not in m.to_pin_dict()
    assert "some_other_field" not in m.to_pin_dict()


def test_meta_factory_create():
    mf = MetaFactory()
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_file = f"{tmp_dir}/some_name"
        with open(tmp_file, "wb") as f:
            f.write(b"test")

        kwargs = {
            "title": "some title",
            "description": "some description",
            "user": {},
            "type": "csv",
            "name": "some_name",
        }

        meta = mf.create(tmp_dir, tmp_file, **kwargs)

        # test that kwargs are passed through ----
        for k, v in kwargs.items():
            assert getattr(meta, k) == v

        # test calculated fields ----
        # TODO(compat): should append suffix to name attr (like in R pins)?
        #       otherwise, will break cross compat?
        assert meta.file == "some_name"
        assert meta.file_size == 4


def test_meta_factory_read_yaml_roundtrip(meta):
    pin_yaml = meta.to_pin_yaml()

    mf = MetaFactory()
    meta2 = mf.read_pin_yaml(StringIO(pin_yaml), meta.name, meta.version)

    assert meta == meta2


def test_meta_factory_roundtrip_unknown(meta):
    meta_dict = meta.to_pin_dict()
    meta_dict["some_other_field"] = 1

    pin_yaml = yaml.dump(meta_dict)

    mf = MetaFactory()

    meta2 = mf.read_pin_yaml(StringIO(pin_yaml), meta.name, meta.version)

    assert meta2 == meta
    assert meta2.some_other_field == 1



import humanize
import logging
import os
import time
import shutil
import urllib.parse

from fsspec.implementations.cached import SimpleCacheFileSystem
from fsspec import register_implementation
from pathlib import Path

from .config import get_cache_dir
from .utils import inform, hash_name

_log = logging.getLogger(__name__)


# used if needed to preserve board path structure in the cache
PLACEHOLDER_VERSION = "v"
PLACEHOLDER_FILE = "file"


def touch_access_time(path, access_time: "float | None" = None, strict=True):
    """Update access time of file.

    Returns the new access time.
    """

    if access_time is None:
        access_time = time.time()

    p = Path(path)

    if not p.exists() and not strict:
        p.touch()

    stat = p.stat()
    os.utime(path, (access_time, stat.st_mtime))

    return access_time


def protocol_to_string(protocol):
    if isinstance(protocol, str):
        return protocol

    return protocol[0]


def prefix_cache(fs, board_base_path):
    if isinstance(fs, str):
        proto_name = fs
    else:
        proto_name = protocol_to_string(fs.protocol)
    base_hash = hash_name(board_base_path, False)

    return f"{proto_name}_{base_hash}"


class HashMapper:
    def __init__(self, hash_prefix):
        self.hash_prefix = hash_prefix

    def __call__(self, path: str) -> str:
        if self.hash_prefix is not None:
            # optionally make the name relative to a parent path
            # using the hash of parent path as a prefix, to flatten a bit
            hash = Path(path).relative_to(Path(self.hash_prefix))
            return hash

        else:
            raise NotImplementedError()


class PinsAccessTimeCacheMapper:
    def __init__(self, hash_prefix):
        self.hash_prefix = hash_prefix

    def __call__(self, path):
        # hash full path, and put anything after the final / at the end, just
        # to make it easier to browse.
        # this has
        base_name = hash_name(path, False)
        suffix = Path(path).name
        return f"{base_name}_{suffix}"


class PinsRscCacheMapper:
    """Modifies the PinsCache to allow hash_prefix to be an RSC server url.

    Note that this class also modifies the first / in a path to be a -, so that
    pin contents will not be put into subdirectories, for e.g. michael/mtcars/data.txt.
    """

    def __init__(self, hash_prefix):
        self.hash_prefix = hash_prefix

    def __call__(self, path):
        # the main change in this function is that, for same_name, it returns
        # the full path
        # change pin path of form <user>/<content> to <user>+<content>
        hash = path.replace("/", "+", 1)
        return hash


class PinsCache(SimpleCacheFileSystem):
    protocol = "pinscache"

    def __init__(self, *args, hash_prefix=None, mapper=HashMapper, **kwargs):
        super().__init__(*args, **kwargs)
        self.hash_prefix = hash_prefix
        self._mapper = mapper(hash_prefix)

    def hash_name(self, path, *args, **kwargs):
        return self._mapper(path)

    def _open(self, path, *args, **kwargs):
        # For some reason, the open method of SimpleCacheFileSystem doesn't
        # call _make_local_details, so we need to patch in here.
        # Note that methods like .cat() do call it. Other Caches don't have this issue.
        path = self._strip_protocol(path)
        self._make_local_details(path)

        return super()._open(path, *args, **kwargs)

    def _make_local_details(self, path):
        # modifies method to create any parent directories needed by the cached file
        # note that this is called in ._open(), at the point it's known the file
        # will be cached
        fn = super()._make_local_details(path)
        _log.info(f"cache file: {fn}")
        Path(fn).parent.mkdir(parents=True, exist_ok=True)

        return fn

    # same as upstream, brought in to preserve backwards compatibility
    def _check_file(self, path):
        self._check_cache()
        sha = self._mapper(path)
        for storage in self.storage:
            fn = os.path.join(storage, sha)
            if os.path.exists(fn):
                return fn


class PinsUrlCache(PinsCache):
    protocol = "pinsurlcache"

    def hash_name(self, path, same_name):
        # strip final arg from path
        # note that R pins uses fs::path_file, and I'm not sure exactly how it
        # behaves for the many forms url paths can take.
        # e.g. path_file(.../extdata/) -> extdata
        # e.g. path_file(.../extdata?123) -> extdata?123
        path_parts = urllib.parse.urlparse(path)[2]

        # strip off final whitespace and / (if it exists)
        # TODO(compat): python pins currently not keeping query part of url
        final_part = path_parts.rstrip().rstrip("/").rsplit("/", 1)[-1]

        # TODO: what happens in R pins if no final part?
        if final_part == "":
            final_part = PLACEHOLDER_FILE

        # hash url
        prefix = hash_name(path, False)

        # note that we include an extra version folder, so it conforms with
        # pin board path form: <board_path>/<pin_name>/<version_name>/<file>
        proto_name = protocol_to_string(self.fs.protocol)
        full_prefix = "_".join([proto_name, prefix])
        return str(Path(full_prefix) / PLACEHOLDER_VERSION / final_part)


class PinsAccessTimeCache(SimpleCacheFileSystem):
    name = "pinsaccesstimecache"

    def __init__(
        self, *args, hash_prefix=None, mapper=PinsAccessTimeCacheMapper, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.hash_prefix = hash_prefix
        self._mapper = mapper(hash_prefix)

    def hash_name(self, path, *args, **kwargs):
        return self._mapper(path)

    def _open(self, path, mode="rb", **kwargs):
        f = super()._open(path, mode=mode, **kwargs)
        fn = self._check_file(path)

        if fn is None:
            raise ValueError(
                f"Cached file should exist for path, but none found: {path}"
            )

        touch_access_time(fn)

        return f

    # same as upstream, brought in to preserve backwards compatibility
    def _check_file(self, path):
        self._check_cache()
        sha = self._mapper(path)
        for storage in self.storage:
            fn = os.path.join(storage, sha)
            if os.path.exists(fn):
                return fn


class CachePruner:
    """Prunes the cache directory, across multiple boards.

    Note
    ----

    `pins` assumes that all boards cache using these rules:

    * path structure: `<cache_root>/<board_hash>/<pin>/<version>`.
    * each version has a data.txt file in it.
    """

    meta_path = "data.txt"

    def __init__(self, cache_dir: "str | Path"):
        self.cache_dir = Path(cache_dir)

    def versions(self) -> "iter[Path]":
        for p_version in self.cache_dir.glob("*/*"):
            if p_version.is_dir() and (p_version / self.meta_path).exists():
                yield p_version

    def should_prune_version(self, days, path: "str | Path"):
        path = Path(path)

        expiry_time_sec = days * 60 * 60 * 24
        prune_before = time.time() - expiry_time_sec

        p_meta = path / self.meta_path

        if not p_meta.exists():
            raise FileNotFoundError(f"No metadata file: {p_meta.absolute()}")

        access_time = p_meta.stat().st_atime
        return access_time < prune_before

    def old_versions(self, days):
        return [p for p in self.versions() if self.should_prune_version(days, p)]

    def prune(self, days=30):
        to_prune = self.old_versions(days)
        size = sum(map(disk_usage, to_prune))

        # TODO: clean this up, general approach to prompting
        confirmed = prompt_cache_prune(to_prune, size)
        if confirmed:
            for path in to_prune:
                delete_version(to_prune)

        _log.info("Skipping cache deletion")


def delete_version(path: "str | Path"):
    path = Path(path)
    shutil.rmtree(str(path.absolute()))


def disk_usage(path):
    return sum(p.stat().st_size for p in path.glob("**/*") if p.is_file() or p.is_dir())


def prompt_cache_prune(to_prune, size) -> bool:
    _log.info(f"Pruning items: {to_prune}")
    human_size = humanize.naturalsize(size, binary=True)
    resp = input(
        f"Delete {len(to_prune)} pin versions, freeing {human_size}?"
        "\n1: Yes"
        "\n2: No"
        "\n\nSelection: "
    )
    return resp == "1"


def cache_info():
    cache_root = get_cache_dir()

    cache_boards = list(Path(cache_root).glob("*"))

    print(f"Cache info: {cache_root}")
    for p_board in cache_boards:
        du = disk_usage(p_board)
        human_size = humanize.naturalsize(du, binary=True)
        rel_path = p_board.relative_to(cache_root)
        print(f"* {rel_path}: {human_size}")


def cache_prune(days=30, cache_root=None, prompt=True):
    if cache_root is None:
        cache_root = get_cache_dir()

    final_delete = []
    for p_board in Path(cache_root).glob("*"):
        pruner = CachePruner(p_board)
        final_delete.extend(pruner.old_versions(days))

    size = sum(map(disk_usage, final_delete))

    if not final_delete:
        inform(_log, "No stale pins found")

    if prompt:
        confirmed = prompt_cache_prune(final_delete, size)
    else:
        confirmed = True

    if confirmed:
        inform(_log, "Deleting pins from cache.")
        for p in final_delete:
            delete_version(p)
    else:
        inform(_log, "Skipping deletion of pins from cache.")


# TODO: swap to use entrypoint
register_implementation("pinscache", PinsCache)



from importlib_resources import files as _files

sources = {
    "mtcars": _files("pins") / "data/mtcars.csv",
}


def __dir__():
    return list(sources)


def __getattr__(k):
    import pandas as pd

    f_path = sources.get("mtcars")

    return pd.read_csv(f_path)



import fsspec
import os
import tempfile

from .boards import BaseBoard, BoardRsConnect, BoardManual
from .cache import PinsCache, PinsRscCacheMapper, PinsAccessTimeCache, prefix_cache
from .config import get_data_dir, get_cache_dir


class DEFAULT:
    pass


# Representing constructors ===================================================

# Note that this is not a constructor, but a function to represent them.
def board_deparse(board: BaseBoard):
    """Return a representation of how a board could be reconstructed.

    Note that this function does not try to represent the exact arguments used
    to construct a board, but key pieces (like the path to the board). You may
    need to specify environment variables with API keys to complete the connection.

    Parameters
    ----------
    board:
        A pins board to be represented.

    Examples
    --------

    The example below deparses a board connected to Posit Connect.

    >>> board_deparse(board_connect(server_url="http://example.com", api_key="xxx"))
    "board_connect(server_url='http://example.com')"

    Note that the deparsing a Posit Connect board does not keep the api_key,
    which is sensitive information. In this case, you can set the CONNECT_API_KEY
    environment variable to connect.

    Below is an example of representing a board connected to a local folder.

    >>> board_deparse(board_folder("a/b/c"))
    "board_folder('a/b/c')"

    >>> board_deparse(board_folder(path="a/b/c", allow_pickle_read=True))
    "board_folder('a/b/c', allow_pickle_read=True)"
    """
    if board.allow_pickle_read is not None:
        allow_pickle = f", allow_pickle_read={repr(board.allow_pickle_read)}"
    else:
        allow_pickle = ""

    prot = board.fs.protocol

    if prot == "rsc":
        url = board.fs.api.server_url
        return f"board_connect(server_url={repr(url)}{allow_pickle})"
    elif prot in ["file", ("file", "local")]:
        return f"board_folder({repr(board.board)}{allow_pickle})"
    elif prot in [["s3", "s3a"], ("s3", "s3a")]:
        return f"board_s3({repr(board.board)}{allow_pickle})"
    elif prot == "abfs":
        return f"board_azure({repr(board.board)}{allow_pickle})"
    elif prot == ("gcs", "gs"):
        return f"board_gcs({repr(board.board)}{allow_pickle})"
    elif prot == "http":
        return f"board_url({repr(board.board)}, {board.pin_paths}{allow_pickle})"
    else:
        raise NotImplementedError(
            f"board deparsing currently not supported for protocol: {prot}"
        )


# Board constructors ==========================================================
# note that libraries not used by board classes above are imported within these
# functions. may be worth moving these funcs into their own module.


def board(
    protocol: str,
    path: str = "",
    versioned: bool = True,
    cache: "type[DEFAULT] | None" = DEFAULT,
    allow_pickle_read=None,
    storage_options: "dict | None" = None,
    board_factory: "callable | type[BaseBoard] | None" = None,
):
    """General function for constructing a pins board.

    Note that this is a lower-level function. For most use cases, use a more specific
    function like [](`~pins.board_local`), or [](`~pins.board_s3`).

    Parameters
    ----------
    protocol: str
        File system protocol. E.g. file, s3, github, rsc (for Posit Connect).
        See `fsspec.filesystem` for more information.
    path: str
        A base path the board should use. For example, the directory the board lives in,
        or the path to its S3 bucket.
    versioned: bool
        Whether or not pins should be versioned.
    cache: optional, type[DEFAULT]
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.
    storage_options:
        Additional options passed to the underlying filesystem created by
        `fsspec.filesystem`.
    board_factory:
        An optional board class to use as the constructor.

    Notes
    -----
    Many fsspec implementations of filesystems cache the searching of files, which may
    cause you to not see pins saved by other people. Disable this on these file systems
    with `storage_options = {"listings_expiry_time": 0}` on S3, or `{"cache_timeout": 0}`
    on Google Cloud Storage.

    """

    if storage_options is None:
        storage_options = {}

    # TODO: at this point should just manually construct the rsc board directly
    # from board_rsconnect...
    if protocol == "rsc":
        # TODO: register RsConnectFs with fsspec
        from pins.rsconnect.fs import RsConnectFs

        fs = RsConnectFs(**storage_options)

    else:
        fs = fsspec.filesystem(protocol, **storage_options)

    # wrap fs in cache ----

    if cache is DEFAULT:
        base_cache_dir = get_cache_dir()

        # manually create a subdirectory for rsc server
        if protocol == "rsc":
            # ensures each server_url is its own cache directory
            hash_prefix = storage_options["server_url"]
            board_cache = prefix_cache(fs, hash_prefix)
            cache_dir = os.path.join(base_cache_dir, board_cache)

            fs = PinsCache(
                cache_storage=cache_dir,
                fs=fs,
                hash_prefix=hash_prefix,
                same_names=True,
                mapper=PinsRscCacheMapper,
            )
        else:
            # ensures each subdir path is its own cache directory
            board_cache = prefix_cache(fs, path)
            cache_dir = os.path.join(base_cache_dir, board_cache)

            fs = PinsCache(
                cache_storage=cache_dir, fs=fs, hash_prefix=path, same_names=True
            )
    elif cache is None:
        pass
    else:
        NotImplemented("Can't currently pass own cache object.")

    # construct board ----

    pickle_kwargs = {"allow_pickle_read": allow_pickle_read}
    # TODO: should use a registry or something
    if board_factory is not None:
        board = board_factory(path, fs, versioned, **pickle_kwargs)
    elif protocol == "rsc":
        board = BoardRsConnect(path, fs, versioned, **pickle_kwargs)
    else:
        board = BaseBoard(path, fs, versioned, **pickle_kwargs)
    return board


# TODO(#31): change file boards to unversioned once implemented


def board_folder(path: str, versioned=True, allow_pickle_read=None):
    """Use a local folder as a board.

    Parameters
    ----------
    path:
        The folder that will hold the board.
    versioned:
        Whether or not pins should be versioned.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.
    """

    return board(
        "file", path, versioned, cache=None, allow_pickle_read=allow_pickle_read
    )


def board_temp(versioned=True, allow_pickle_read=None):
    """Use a local temporary directory as a board.

    Parameters
    ----------
    versioned:
        Whether or not pins should be versioned.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.
    """

    tmp_dir = tempfile.TemporaryDirectory()

    board_obj = board(
        "file", tmp_dir.name, versioned, cache=None, allow_pickle_read=allow_pickle_read
    )

    # TODO: this is necessary to ensure the temporary directory dir persists.
    # without maintaining a reference to it, it could be deleted after this
    # function returns
    board_obj.__tmp_dir = tmp_dir

    return board_obj


def board_local(versioned=True, allow_pickle_read=None):
    """Use a local folder as a board.

    Parameters
    ----------
    versioned:
        Whether or not pins should be versioned.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.
    """
    path = get_data_dir()

    return board(
        "file", path, versioned, cache=None, allow_pickle_read=allow_pickle_read
    )


def board_github(
    org,
    repo,
    path="",
    token=None,
    versioned=True,
    cache=DEFAULT,
    allow_pickle_read=None,
):
    """Create a board to read and write pins from GitHub.

    Parameters
    ----------
    org:
        Name of the GitHub org (e.g. user account).
    repo:
        Name of the repo.
    path:
        A subfolder in the GitHub repo holding the board.
    token:
        An optional GitHub token.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.


    Notes
    -----
    This board is read only.


    Examples
    --------

    >>> import pytest; pytest.skip()

    >>> import os
    >>> board = board_github("machow", "pins-python", "pins/tests/pins-compat")
    >>> board.pin_list()
    ['df_arrow', 'df_csv', 'df_rds', 'df_unversioned']

    >>> board.pin_read("df_csv")
       x  y  z
    0  1  a  3
    1  2  b  4


    """

    return board(
        "github",
        path,
        versioned,
        cache,
        allow_pickle_read=allow_pickle_read,
        storage_options={"org": org, "repo": repo, "listings_expiry_time": 0},
    )


def board_urls(*args, **kwargs):
    """DEPRECATED: This board has been renamed to board_url."""
    from .utils import warn_deprecated

    warn_deprecated(
        "board_urls has been renamed to board_url. Please use board_url instead."
    )

    return board_url(*args, **kwargs)


def board_url(path: str, pin_paths: dict, cache=DEFAULT, allow_pickle_read=None):
    """Create a board from individual URLs.

    Parameters
    ----------
    path:
        A base URL to prefix all individual pin URLs with.
    pin_paths: Mapping
        A dictionary mapping pin name to pin URL.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.


    Examples
    --------

    ```python
    github_raw = "https://raw.githubusercontent.com/rstudio/pins-python/main/pins/tests/pins-compat"
    pin_paths = {
      "df_csv": "df_csv/20220214T163720Z-9bfad/",
      "df_arrow": "df_arrow/20220214T163720Z-ad0c1/",
      }
    board = board_url(github_raw, pin_paths)
    board.pin_list()
    ```
    ```
    ['df_csv', 'df_arrow']
    ```
    """

    # TODO(compat): R pins' version is named board_url (no s)
    if cache is DEFAULT:
        # copied from board(). this ensures that paths in cache have the form:
        # <full_path_hash>/<version_placeholder>/<file_name>
        cache_dir = get_cache_dir()
        sub_dir = prefix_cache("http", path)
        sub_cache = f"{cache_dir}/{sub_dir}"
        fs = PinsAccessTimeCache(
            target_protocol="http", cache_storage=sub_cache, same_names=False
        )
    else:
        raise NotImplementedError("Can't currently pass own cache object")

    return BoardManual(
        path,
        fs,
        versioned=False,
        allow_pickle_read=allow_pickle_read,
        pin_paths=pin_paths,
    )


def board_connect(
    server_url=None, versioned=True, api_key=None, cache=DEFAULT, allow_pickle_read=None
):
    """Create a board to read and write pins from a Posit Connect server.

    Parameters
    ----------
    server_url:
        URL to the Posit Connect server.
    versioned:
        Whether or not pins should be versioned.
    api_key:
        API key for server. If not specified, pins will attempt to read it from
        `CONNECT_API_KEY` environment variable.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.


    Examples
    --------
    Pins will automatically look for the `CONNECT_SERVER` and `CONNECT_API_KEY` environment variables:

    ```python
    # where environment vars CONNECT_SERVER and CONNECT_API_KEY are set
    board = board_connect()
    ```

    Or use the [dotenv](https://saurabh-kumar.com/python-dotenv/) package to load other environment variable names from a `.env` file:

    ```python
    import os
    from dotenv import load_dotenv, find_dotenv
    load_dotenv(find_dotenv())

    api_key = os.getenv("MY_API_KEY")
    server_url = os.getenv("MY_CONNECT_URL")
    board = board_connect(server_url=server_url, api_key=api_key)
    ```

    In order to read a public pin, use `board_url()` with the public pin URL:

    ```python
    # for a pin at https://connect.rstudioservices.com/content/3004/
    board = board_url(
      "https://connect.rstudioservices.com/content",
      {"my_df": "3004/"}
    )
    board.pin_read("my_df")
    ```

    See Also
    --------
    [](`~pins.board_url`) : Board for connecting to individual pins, using a URL or path.

    """

    # TODO: api_key can be passed in to underlying RscApi, equiv to R's manual mode
    # TODO: otherwise, CONNECT_API_KEY and CONNECT_SERVER env vars should also work
    if server_url is None:
        # TODO: this should be inside the api class
        server_url = os.environ.get("CONNECT_SERVER")

    kwargs = dict(server_url=server_url, api_key=api_key)
    return board(
        "rsc", None, versioned, cache, allow_pickle_read, storage_options=kwargs
    )


board_rsconnect = board_connect


def board_s3(path, versioned=True, cache=DEFAULT, allow_pickle_read=None):
    """Create a board to read and write pins from an AWS S3 bucket folder.

    Parameters
    ----------
    path:
        Path of form `<bucket_name>/<optional>/<subdirectory>`.
    versioned:
        Whether or not pins should be versioned.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.

    Notes
    -----
    The s3 board uses the fsspec library (s3fs) to handle interacting with AWS S3.
    In order to authenticate, set the `AWS_ACCESS_KEY_ID`, `AWS_SECRET_ACCESS_KEY`,
    and (optionally) `AWS_REGION` environment variables.

    See <https://github.com/fsspec/s3fs>

    """
    # TODO: user should be able to specify storage options here?

    opts = {"listings_expiry_time": 0}
    return board("s3", path, versioned, cache, allow_pickle_read, storage_options=opts)


def board_gcs(path, versioned=True, cache=DEFAULT, allow_pickle_read=None):
    """Create a board to read and write pins from a Google Cloud Storage bucket folder.

    Parameters
    ----------
    path:
        Path of form `<bucket_name>/<optional>/<subdirectory>`.
    versioned:
        Whether or not pins should be versioned.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.

    Notes
    -----
    The gcs board uses the fsspec library (gcsfs) to handle interacting with
    Google Cloud Storage. Currently, its default mode of authentication
    is supported.

    See <https://gcsfs.readthedocs.io/#credentials>
    """

    # GCSFS uses a different name for listings_expiry_time, and then
    # fixes it under the hood
    opts = {"cache_timeout": 0}
    return board("gcs", path, versioned, cache, allow_pickle_read, storage_options=opts)


def board_azure(path, versioned=True, cache=DEFAULT, allow_pickle_read=None):
    """Create a board to read and write pins from an Azure Datalake Filesystem folder.

    Parameters
    ----------
    path:
        Path of form `<bucket_name>/<optional>/<subdirectory>`.
    versioned:
        Whether or not pins should be versioned.
    cache:
        Whether to use a cache. By default, pins attempts to select the right cache
        directory, given your filesystem. If `None` is passed, then no cache will be
        used. You can set the cache using the `PINS_CACHE_DIR` environment variable.
    allow_pickle_read: optional, bool
        Whether to allow reading pins that use the pickle protocol. Pickles are unsafe,
        and can execute arbitrary code. Only allow reading pickles if you trust the
        board to execute Python code on your computer.

        You can enable reading pickles by setting this to `True`, or by setting the
        environment variable `PINS_ALLOW_PICKLE_READ`. If both are set, this argument
        takes precedence.

    Notes
    -----
    The azure board uses the fsspec library (adlfs) to handle interacting with
    Azure Datalake Filesystem (abfs). Currently, its default mode of authentication
    is supported.

    See <https://github.com/fsspec/adlfs>
    """

    opts = {"use_listings_cache": False}
    return board(
        "abfs", path, versioned, cache, allow_pickle_read, storage_options=opts
    )
